Package: warp
Title: Group Dates
Version: 0.2.1
Authors@R: c(
    person("Davis", "Vaughan", , "davis@posit.co", role = c("aut", "cre")),
    person("Posit Software, PBC", role = c("cph", "fnd"))
  )
Description: Tooling to group dates by a variety of periods including:
    yearly, monthly, by second, by week of the month, and more.  The
    groups are defined in such a way that they also represent the distance
    between dates in terms of the period. This extracts valuable
    information that can be used in further calculations that rely on a
    specific temporal spacing between observations.
License: MIT + file LICENSE
URL: https://github.com/DavisVaughan/warp,
        https://davisvaughan.github.io/warp/
BugReports: https://github.com/DavisVaughan/warp/issues
Depends: R (>= 3.2)
Suggests: covr, knitr, rmarkdown, testthat (>= 3.0.0)
VignetteBuilder: knitr
Config/Needs/website: tidyverse/tidytemplate
Encoding: UTF-8
Roxygen: list(markdown = TRUE)
RoxygenNote: 7.2.3
Author: Davis Vaughan [aut, cre],
  Posit Software, PBC [cph, fnd]
Maintainer: Davis Vaughan <davis@posit.co>
Built: R 4.4.0; x86_64-apple-darwin18.7.0; 2024-04-28 12:32:55 UTC; unix
