Package: warp
Title: Group Dates
Version: 0.2.0
Authors@R: 
    c(person(given = "Davis",
             family = "Vaughan",
             role = c("aut", "cre"),
             email = "davis@rstudio.com"),
      person(given = "RStudio",
             role = "cph"))
Description: Tooling to group dates by a variety of periods
    including: yearly, monthly, by second, by week of the month, and more.
    The groups are defined in such a way that they also represent the
    distance between dates in terms of the period. This extracts valuable
    information that can be used in further calculations that rely on a
    specific temporal spacing between observations.
License: MIT + file LICENSE
URL: https://github.com/DavisVaughan/warp
BugReports: https://github.com/DavisVaughan/warp/issues
Depends: R (>= 3.2)
Suggests: covr, testthat (>= 2.1.0), knitr, rmarkdown
Encoding: UTF-8
LazyData: true
RoxygenNote: 7.1.1
VignetteBuilder: knitr
NeedsCompilation: yes
Packaged: 2020-10-21 17:22:54 UTC; davis
Author: Davis Vaughan [aut, cre],
  RStudio [cph]
Maintainer: Davis Vaughan <davis@rstudio.com>
Repository: CRAN
Date/Publication: 2020-10-21 18:00:02 UTC
Built: R 4.1.0; x86_64-w64-mingw32; 2021-05-21 22:19:32 UTC; windows
Archs: x64
